package ru.ispras.telecomtesk.tlsv13.HandshakeMessages;

import java.security.*;
import java.security.cert.Certificate;
import java.util.ArrayList;

import javax.crypto.spec.SecretKeySpec;
import javax.xml.bind.DatatypeConverter;

import ru.ispras.telecomtesk.tlsv13.ConvertTypes;
import ru.ispras.telecomtesk.tlsv13.DebugInfo;
import ru.ispras.telecomtesk.tlsv13.TlsKeystore;
import ru.ispras.telecomtesk.tlsv13.HandshakeMessages.TLSExtensions.TLSExtensionsList;
import ru.ispras.telecomtesk.tlsv13.TLSAlgorithms.SignatureScheme;
import ru.ispras.telecomtesk.tlsv13.TLSAlgorithms.TLS13SpecValues;
import ru.ispras.telecomtesk.tlsv13.TLSAlgorithms.HashAlgorithm;
import ru.ispras.telecomtesk.tlsv13.TLSAlgorithms.MACAlgorithm;
import ru.ispras.telecomtesk.tlsv13.TLSAlgorithms.SignatureAlgorithm;
import ru.ispras.telecomtesk.tlsv13.TLSState.ConnectionRole;
import ru.ispras.telecomtesk.tlsv13.TLSState.TLSConnection;
import ru.ispras.telecomtesk.tlsv13.TLSState.TLSKeyDerivation;

public class CertificateVerify extends HandshakeMessage
{
	public SignatureScheme alg;
	private byte[] signature;
	private boolean set_signature;
	public int error = 0;
    
	
	public CertificateVerify(SignatureScheme sah) {
		msg_type = 15; 
		alg = sah;
		signature = new byte[0];
		set_signature = false;
	}
	public CertificateVerify(SignatureScheme sah, byte[] signature) {
		this(sah);
		setSignature(signature);
	}
	
	public void setSignature(byte[] signature) {
		if(signature != null) {
			this.signature = signature;
			set_signature = true;
		}
	}
	
	public byte[] getSignature() {
		return this.signature;
	}
	
	public void calcSignature(TLSConnection con, ArrayList msglist) throws Exception {
		Signature sig;
		String hash_name, sig_name, alg_name;
		MACAlgorithm malg;
		Object curmsg;
		ArrayList<HandshakeMessage> handshakeContext = new ArrayList<HandshakeMessage>();

		for(int i=0;i<msglist.size();i++) {
			curmsg = msglist.get(i);
			if(curmsg instanceof HandshakeMessage) {
				if(curmsg instanceof HandshakeUndefinedType) continue;  // TEST message type, not include in Hash calculating
				handshakeContext.add((HandshakeMessage)curmsg);
			}
		}

		malg = con.getSecurityParams().mac_alg;
		byte[] hashContentPrefix = con.localRole == ConnectionRole.CLIENT ? TLS13SpecValues.certificateVerifyClientPrefix : TLS13SpecValues.certificateVerifyServerPrefix;
		byte[] hashHandshakeContext = TLSKeyDerivation.TLS_TranscriptHash(malg, handshakeContext);
		byte[] hashContent = new byte[hashContentPrefix.length + hashHandshakeContext.length];
		System.arraycopy(hashContentPrefix, 0, hashContent, 0, hashContentPrefix.length);
		System.arraycopy(hashHandshakeContext, 0, hashContent, hashContentPrefix.length, hashHandshakeContext.length);
		
		alg_name = this.alg.getHashName() + "with" + this.alg.getSigName();
			
		Certificate localCert = con.getHandshakeParams().localCert;
		TlsKeystore.getSigAlg(localCert);
		KeyStore localks = null;
		String localKeystorePwd = "";
		PrivateKey localPrivateKey = TlsKeystore.getPrivateKeyCert(localks, localCert, localKeystorePwd);
		sig = Signature.getInstance(alg_name,"BC");
		sig.initSign(localPrivateKey);
		sig.update(hashContent);
		this.setSignature(sig.sign());
	}

	public boolean verifySignature(TLSConnection con, byte[] signat) throws Exception {
		Signature sig;
		String alg_name;
		MACAlgorithm malg;
		Object curmsg;
		boolean res = false;
		
		ArrayList<HandshakeMessage> handshakeContext = new ArrayList<HandshakeMessage>();
		ArrayList msglist = con.getHandshakeParams().sessionMsgList;

		for(int i=0;i<msglist.size();i++) {
			curmsg = msglist.get(i);
			if(curmsg instanceof HandshakeMessage) {
				if(curmsg instanceof HandshakeUndefinedType) continue;  // TEST message type, not include in Hash calculating
				handshakeContext.add((HandshakeMessage)curmsg);
			}
		}

		malg = con.getSecurityParams().mac_alg;
		byte[] hashContentPrefix = con.localRole == ConnectionRole.SERVER ? TLS13SpecValues.certificateVerifyClientPrefix : TLS13SpecValues.certificateVerifyServerPrefix;
		byte[] hashHandshakeContext = TLSKeyDerivation.TLS_TranscriptHash(malg, handshakeContext);
		byte[] hashContent = new byte[hashContentPrefix.length + hashHandshakeContext.length];
		System.arraycopy(hashContentPrefix, 0, hashContent, 0, hashContentPrefix.length);
		System.arraycopy(hashHandshakeContext, 0, hashContent, hashContentPrefix.length, hashHandshakeContext.length);
		
		alg_name = this.alg.getHashName() + "with" + this.alg.getSigName();
			
		sig = Signature.getInstance(alg_name);
		sig.initVerify(con.getHandshakeParams().remoteCert);
		DebugInfo.printDump(5,hashContent,"verifySignature:hashContent:");
		DebugInfo.printDump(5,signat,"verifySignature:signature:");
		sig.update(hashContent);
		res = sig.verify(signat);
		
		DebugInfo.printMessage(3, "CertificateVerify::verifySignature: return " + res);
		return res;
	}

	// 4 bytes + 2 byte (SignatureScheme) + 2 byte (signature length) + length[signature]
	@Override
	public int getSize() {
		return (8 + signature.length);
	}


	@Override
	public String toString() {
		String res = "HandshakeMessage: type CertificateVerify, length " + length;
		if(net_msg != null)
			res += ", net_msg: " + net_msg.length;
		res += "\n      SignatureScheme: " + alg;
		res += "\n      signature length: " + signature.length;
		return res;
	}


	@Override
	public void print() {
		DebugInfo.printMessage(1,this.toString());
		DebugInfo.printDump(1,this.signature,"signature:");
	}

	@Override
	public int toByteArray() {
		if(isLengthSet == false) setLength();
		// create byte array
		byte[] ba = new byte[getSize()];
		int res = 0, i;
		byte[] ba_len = ConvertTypes.hton(length);
		ba[0] = msg_type;
		System.arraycopy(ba_len, 1, ba, 1, 3);
		i=4;
		ba[i++] = (byte)(alg.getValue() >>> 8);
		ba[i++] = (byte)(alg.getValue());
		ba[i++] = (byte)(signature.length >>> 8);
		ba[i++] = (byte)signature.length;
		System.arraycopy(signature, 0, ba, i, signature.length);

		net_msg = ba;
		return res;
	}
	
    public static HandshakeMessage parseByteArray(byte[] msg) {
    	int siglen = 0;

    	// check header length: 4 bytes + 2 byte (SignatureScheme) + 2 byte (signature length)
    	if(msg.length < 8) {
    		DebugInfo.printMessage(4,"===Error! CertificateVerify::parseByteArray: message length failed");
    		return null;
		}
    	
    	// check signature length
    	siglen = ConvertTypes.ntohs(msg,6); // convert byte[2] to short
    	if(msg.length != siglen + 8) {
    		DebugInfo.printMessage(4,"===Error! CertificateVerify::parseByteArray: signature length wrong");
    		return null;
		}

    	SignatureScheme alg = SignatureScheme.valueOf(ConvertTypes.ntohs(msg,4));

    	byte[] signature = new byte[siglen];
    	System.arraycopy(msg, 8, signature, 0, siglen);
    	
    	return new CertificateVerify(alg, signature);
    }
    	
}
