package ru.ispras.telecomtesk.tlsv13.HandshakeMessages;

import java.io.ByteArrayInputStream;
import ru.ispras.telecomtesk.tlsv13.ConvertTypes;
import ru.ispras.telecomtesk.tlsv13.DebugInfo;
import ru.ispras.telecomtesk.tlsv13.TLSAlgorithms.KeyExchangeMode;

import java.security.cert.CertificateFactory;
import java.security.cert.X509Certificate;
import java.util.ArrayList;

public class HMCertificate extends HandshakeMessage {    // HandshakeMessages Certificate
	public byte[] certificateRequestContext;             // <0..2^8-1>
	public ArrayList<CertificateEntry> certificateList; // <0..2^24-1> 

	public HMCertificate() {
		this(null,null);
	}
	
	public HMCertificate(byte[] context) {
		this(context,null);
	}
	
	public HMCertificate(ArrayList<CertificateEntry> certList) {
		this(null,certList);
	}
	
	public HMCertificate(byte[] context, ArrayList<CertificateEntry> certList) {
		this.msg_type = 11;
		this.certificateRequestContext = context != null ? context : new byte[0];
		this.certificateList = certList != null ? certList : new ArrayList<CertificateEntry>();
	}
	
	public void addCertificate(CertificateEntry ext) {
		certificateList.add(ext);
	}
	
	public void removeCertificate(CertificateEntry ext) {
		certificateList.remove(ext);
	}
	
	public void clearCertificateList() {
		certificateList.clear();
	}

	public CertificateEntry getFirstCertificate() {
		if(!certificateList.isEmpty())
			return certificateList.get(0);
		else
			return null;
	}
	
	public int getExtensionLength() {
		int ext_len = 0;
		for (CertificateEntry ext : certificateList) {
			ext_len += ext.getSize();
		}
		return ext_len;
	}

	@Override
	public int getSize() {
		return (4 + 1 + certificateRequestContext.length + 3 + this.getExtensionLength());
	}
	
	public boolean isCertificateListEmpty() {
		return certificateList.isEmpty();
	}
	
	@Override
	public int toByteArray() {
		int i = 0, len, res = 0;
		if(isLengthSet == false) setLength();
		// create byte array
		len = getSize();
		byte[] ba = new byte[len];
		// write msg_type and full length
		byte[] ba_len = ConvertTypes.hton(length);
		ba[0] = msg_type;
		System.arraycopy(ba_len, 1, ba, 1, 3);
		i=4;
		// write certificateRequestContext
		ba[i++] = (byte)certificateRequestContext.length;
		System.arraycopy(certificateRequestContext, 0, ba, i, certificateRequestContext.length);
		i += certificateRequestContext.length;
		// write certificate_list
		ba_len = ConvertTypes.hton(getExtensionLength()); // return byte[4] 
		System.arraycopy(ba_len, 1, ba, i, 3);
		i+=3;
		byte[] extRaw;
		for (CertificateEntry ext : certificateList) {
			extRaw = ext.toByteArray();
			System.arraycopy(extRaw, 0, ba, i, extRaw.length);
			i += extRaw.length;
		}
		
		net_msg = ba;
		return res;
	}

	public String toString() {
		String res = "HandshakeMessage.Certificate: \n" 
				+ "certificateRequestContext length " + certificateRequestContext.length + 
				", certificateList size " + certificateList.size();
			X509Certificate curcert;
			for(CertificateEntry cert : certificateList) {
					res += "\n   Certificate " + ": length " + cert.getData().length;
					try {
						CertificateFactory factory = CertificateFactory.getInstance("X509");
						curcert = (X509Certificate)factory.generateCertificate( new ByteArrayInputStream(cert.getData()) );
						res += ", X509, Key type: " + curcert.getPublicKey().getAlgorithm();
						res += ", SigAlgName: " + curcert.getSigAlgName();
					}
					catch (Exception err) {
						res += ", not X509 type";
					}
					res += "\n   number of extensions " + cert.getExtensions().size();
			}

		return res;
	}
	
	public void print() {
		DebugInfo.printMessage(1,"\nHandshake Certificate Message:");
		DebugInfo.printDump(1,certificateRequestContext,"CertificateRequestContext:");
		DebugInfo.printMessage(1,"certificateList size " + certificateList.size());
		for(CertificateEntry cert : certificateList) 
			cert.print();
	}

    public static HandshakeMessage parseByteArray(byte[] msg) {
    	int pos = 4, curlen;

    	// parse CertificateRequestContext
    	if(msg.length <= pos) {
    		DebugInfo.printMessage(4,"===Error! HMCertificate::parseByteArray: CertificateRequestContext length failed");
    		return null;
		}
    	curlen = msg[pos++];
    	if(msg.length - pos < curlen) {
    		DebugInfo.printMessage(4,"===Error! HMCertificate::parseByteArray: CertificateRequestContext length wrong");
    		return null;
		}
    	byte[] context = new byte[curlen];
		System.arraycopy(msg, pos, context, 0, curlen);
		pos += curlen;
    	
    	return new HMCertificate(context, parseCertificateEntryList(msg, pos));
    }

    // CertificateEntry certificate_list<0..2^24-1>;
    public static ArrayList<CertificateEntry> parseCertificateEntryList(byte[] msg, int pos) {
    	ArrayList<CertificateEntry> certList = new ArrayList<CertificateEntry>();

    	if((pos >= 0)&&(msg.length >= pos + 3)) {
    		int extlen = msg.length - pos;
       		extlen = ConvertTypes.ntoh(msg,pos,3);
       		pos += 3;
       		if(extlen == msg.length - pos) {
       			CertificateEntry curext;
               	int len;
                	
           		while(extlen > 0) {
           			curext = CertificateEntry.parseByteArray(msg, pos);
           			if(curext == null) break;
           			certList.add(curext);
                    	
               		len = curext.getSize();
               		pos += len;
               		extlen -= len;
           		}
       		} else {
           		DebugInfo.printMessage(4,"===Error! HMCertificate::parseCertificateEntryList: CertificateList length wrong");
       		}
   		} else {
       		DebugInfo.printMessage(4,"===Error! HMCertificate::parseCertificateEntryList: CertificateList length failed");
   		}
    		
    	return certList;
    }

	public static boolean needServerCertificate(KeyExchangeMode ke) {
		switch(ke) {
		case DHE:
			return true;
		default:
			return false;
		}
	}
}
