package ru.ispras.telecomtesk.tlsv13.HandshakeMessages.TLSExtensions;

import ru.ispras.telecomtesk.tlsv13.ConvertTypes;
import ru.ispras.telecomtesk.tlsv13.DebugInfo;

public class TLSExtension
{
	public short type;
	public short length;
	protected boolean isLengthSet = false;
	public byte[] data; // 0..2^16-1 bytes
	
	public TLSExtension() {
		type = 0; length = 0;
	}
	public TLSExtension(int type, byte[] data) {
		this.type = (short)type; 
		this.length = (short)(data.length); 
		this.data = data;
	}
	public void setLength(int len) {
		length = (short)len;
		isLengthSet = true;
	}
	public void setLength(short len) {
		length = len;
		isLengthSet = true;
	}
	public void setLength() {
		isLengthSet = true;
		length = (short) (getLengthFieldSize() + getDataSize());
	}

	public int getExtHeaderSize() {
		return 4;
	}

	public int getLengthFieldSize() {
		return 0;
	}

	public int getDataSize() {
		if(data != null)
			return (data.length);
		else
			return 0;
	}

	public TLSExtensionType getExtensionType() {
		return TLSExtensionType.valueOf(this.type);
	}

	public int getSize() {
		return (getExtHeaderSize() + getLengthFieldSize() + getDataSize());
	}

	public void toByteArrayHeader(byte[] msg, int position) {
		msg[position++] = (byte)(type >>> 8);
		msg[position++] = (byte)type;
		msg[position++] = (byte)(length >>> 8);
		msg[position]   = (byte)length;
	}

	public void toByteArrayLengthField(byte[] msg, int position) {
		int len = getLengthFieldSize();
		if(len > 0) {
			int dataLen = getDataSize();
			if(len == 2) {
				msg[position++] = (byte)(dataLen >>> 8);
				msg[position]   = (byte)dataLen;
			} else if (len == 1)
				msg[position]   = (byte)dataLen;			
		}
	}

	public void toByteArrayData(byte[] msg, int position) {
		if(data != null)
			System.arraycopy(data, 0, msg, position, data.length);
	}

	public byte[] toByteArray() {
		byte[] res = new byte[getSize()];
		return toByteArray(res, 0);
	}

	public byte[] toByteArray(byte[] msg, int position) {
		if(isLengthSet == false) setLength();

		if((msg == null) || (position < 0) || (msg.length - position < getSize()))
			return null;

		toByteArrayHeader(msg,position);        // write Extension Header
		position += getExtHeaderSize();
		toByteArrayLengthField(msg,position);   // write Length of data if needed
		position += getLengthFieldSize();
		toByteArrayData(msg,position);          // write data

		return msg;
	}

	public void printType() {
		DebugInfo.printMessage(1,"TLSExtension type: " + TLSExtensionType.toString(type));
	}
	public void printData() {
		DebugInfo.printDump(1,data,"  extension data:");
	}
	public void print() {
		printType();
		printData();
	}

	// HelloMsgType: 1 - ClientHello, 2 - ServerHello, 3 - HelloRetry
	public static TLSExtension parseByteArray(byte[] msg, int pos, int HelloMsgType) {
		TLSExtension curext = null;
		short exttype;
		int len;
		if(msg.length - pos < 4) {
            		DebugInfo.printMessage(4,"===Error! TLSExtension::parseByteArray: Extension header failed");
            		return null;
        	}
        	exttype = ConvertTypes.ntohs(msg,pos);
        	pos+=2;
        	len = ConvertTypes.ntohs(msg,pos);
        	pos+=2;
        		
        	if(msg.length - pos < len) {
            		DebugInfo.printMessage(4,"===Error! TLSExtension::parseByteArray: Extension failed");
            		return null;
        	}
        		
        	switch (exttype) {
				case 10:
					curext = SupportedGroupsExt.parseByteArray(len, msg, pos);
        			break;
        		case 13:
    				curext = SignatureAndHashAlgorithmExt.parseByteArray(len, msg, pos);
        			break;
    			case 41:
    				if(HelloMsgType == 1)
    					curext = PreSharedKeyClientHello.parseByteArray(len, msg, pos);
    				else if (HelloMsgType == 2)
    					curext = PreSharedKeyServerHello.parseByteArray(len, msg, pos);
        			break;
				case 42:
					curext = EarlyDataExt.parseByteArray(len, msg, pos);
        			break;
				case 43:
					curext = SupportedVersionsExt.parseByteArray(len, msg, pos, HelloMsgType);
        			break;
    			case 44:
    				curext = CookieExt.parseByteArray(len, msg, pos);
        			break;
    			case 45:
    				curext = PSKexchangeModesExt.parseByteArray(len, msg, pos);
        			break;
				case 49:
					curext = PostHandshakeAuthExt.parseByteArray(len, msg, pos);
        			break;
        		case 50:
    				curext = SignatureAlgorithmCertExt.parseByteArray(len, msg, pos);
        			break;
    			case 51:
    				if(HelloMsgType == 1)
    					curext = KeyShareClientHello.parseByteArray(len, msg, pos);
    				else if (HelloMsgType == 2)
    					curext = KeyShareServerHello.parseByteArray(len, msg, pos);
    				else if (HelloMsgType == 3)
    					curext = KeyShareHelloRetry.parseByteArray(len, msg, pos);
        			break;
        		case (short)0xff01:
    				curext = RenegotiationExtension.parseByteArray(len, msg, pos);
        			break;
        		default:
                	byte[] extdata = new byte[len];
        			System.arraycopy(msg, pos, extdata, 0, len);
        			curext = new TLSExtension(exttype, extdata);
        	}

		return curext;
	}
}
