package ru.ispras.telecomtesk.tlsv13.TLSState;

import javax.crypto.*;
import javax.crypto.spec.IvParameterSpec;
import javax.crypto.spec.SecretKeySpec;
import java.util.Random;

import ru.ispras.telecomtesk.tlsv13.DebugInfo;
import ru.ispras.telecomtesk.tlsv13.TLSAlgorithms.*;

class TLSStateKeys
{
	byte[] macKey;
	byte[] encKey;
	byte[] fixedIV;
	long sec_num;

	public TLSStateKeys() {
		macKey = null; encKey = null; fixedIV = null;
		sec_num = 0;
	}
	public TLSStateKeys(int mac, int enc, int iv) {
		macKey = new byte[mac]; encKey = new byte[enc]; fixedIV = new byte[iv];
		sec_num = 0;
	}
	public TLSStateKeys(byte[] mac, byte[] enc, byte[] iv) {
		macKey = mac; encKey = enc; fixedIV = iv;
		sec_num = 0;
	}
}


public class TLSState
{
	public byte[] sessionID;
	public TLSSecurityParameters params;
	TLSStateKeys keys;
	public enum PaddingFill {FILL_PAD_LEN, FILL_PAD_ZERO, FILL_PAD_NUM};
	public enum TLSStateKeyType {MAC_KEY, ENC_KEY, FIXEDIV_KEY};
	
	public TLSState(){
		params = new TLSSecurityParameters();
		keys = new TLSStateKeys();
	}
	public TLSState(boolean isServer){
		params = new TLSSecurityParameters();
		keys = new TLSStateKeys();
	}
	public TLSState(byte[] id, TLSSecurityParameters par, TLSStateKeys keys) {
		setSessionID(id);
		params = par; this.keys = keys; 
	}
	public void setKeys(TLSStateKeys keys) {
		this.keys = keys;
	}
	public void setSessionID(byte[] id) {
		if(id != null) {
			sessionID = new byte[id.length];
			System.arraycopy(id, 0, sessionID, 0, id.length);
		}
	}
	public byte[] getSessionID() {
		return sessionID;
	}
	public TLSSecurityParameters getParams() {
		return params;
	}

	public byte[] compress(byte[] data) {
		byte[] res = null;
		
		switch (params.compression_alg) {
			case 0:
				res = data; break;
			default:
				System.out.println("TLSState::compress: Algorithm not supported");
		}
		return res;
	}

	public byte[] encrypt(byte[] data, byte[] padding) throws Exception {
		byte[] res = null;
		SecretKeySpec sk;
		String alg, algmode;
		Random rnd;
		int padlen;
		Cipher cobj;
		
		DebugInfo.printMessage(4,"TLSState::encrypt: cipher_type " + params.cipher_type);
		switch (params.cipher_type) {
			case stream:
				switch (params.bulk_cipher_alg) {
					case bulk_null:
						return data;
					case bulk_rc4:
						alg = "RC4";
						break;
					default:
						System.out.println("TLSState::decrypt: Algorithm not supported");
						return null;
				} 
				cobj = Cipher.getInstance(alg);
				sk = new SecretKeySpec(keys.encKey, alg);
				cobj.init(Cipher.ENCRYPT_MODE, sk);
				res = cobj.doFinal(data); 
				break;

			case block:
				switch (params.bulk_cipher_alg) {
				case bulk_3des:
					alg = "DESede";
					algmode = "DESede/CBC/NoPadding"; break;
				case bulk_aes128:
				case bulk_aes256:
					alg = "AES";
					algmode = "AES/CBC/NoPadding"; break;
				default:
					System.out.println("TLSState::decrypt: Algorithm not supported");
					return null;
				} 
				// init Crypto Object
				sk = new SecretKeySpec(keys.encKey, alg);
				cobj = Cipher.getInstance(algmode);
				rnd = new Random(System.currentTimeMillis());
				byte[] iv = new byte[params.record_iv_length];
				rnd.nextBytes(iv);
				IvParameterSpec paramSpec = new IvParameterSpec(iv);
				cobj.init(Cipher.ENCRYPT_MODE, sk, paramSpec);
				// create padding
				if(padding == null) {
					if(params.block_length != 0) {
						padlen = params.block_length - data.length % params.block_length;
						padding = createPadding(padlen, PaddingFill.FILL_PAD_LEN);
					}
					else
						padding = new byte[0];
				}
				DebugInfo.printDump(4,padding,"padding");
				// encrypt data
				int curpos;
				int enclen = cobj.getOutputSize(data.length + padding.length);
				res = new byte[enclen + params.record_iv_length];
				System.arraycopy(iv, 0, res, 0, params.record_iv_length);
				curpos = cobj.update(data, 0, data.length, res, params.record_iv_length);
				if(padding.length != 0)
					curpos += cobj.update(padding, 0, padding.length, res, params.record_iv_length + curpos);
				cobj.doFinal(res, params.record_iv_length + curpos); 
				break;

			case aead:
			default:
				break;
		}
		
		return res;
	}
	
	public byte[] encrypt(byte[][] data, byte[] padding) throws Exception {
		byte[] res = null;
		SecretKeySpec sk;
		String alg, algmode;
		Random rnd;
		int padlen, curpos, enclen, datalen = 0, i;
		Cipher cobj;

		DebugInfo.printMessage(4,"TLSState::encrypt: cipher_type " + params.cipher_type);
		for(i=0; i<data.length; i++){
			if(data[i] != null)
				datalen += data[i].length;
		}

		switch (params.cipher_type) {
			case stream:
				switch (params.bulk_cipher_alg) {
					case bulk_null:
						res = new byte[datalen];
						for(i=0,curpos=0; i<data.length; i++){
							if(data[i] != null) {
								System.arraycopy(data[i], 0, res, curpos, data[i].length);
								curpos += data[i].length;
							}
						}
						return res;
					case bulk_rc4:
						alg = "RC4";
						break;
					default:
						System.out.println("TLSState::decrypt: Algorithm not supported");
						return null;
				} 
				// init Crypto Object
				cobj = Cipher.getInstance(alg);
				sk = new SecretKeySpec(keys.encKey, alg);
				cobj.init(Cipher.ENCRYPT_MODE, sk);
				// encrypt data
				enclen = cobj.getOutputSize(datalen);
				res = new byte[enclen];
				for(i=0,curpos=0; i<data.length; i++){
					if(data[i] != null)
						curpos += cobj.update(data[i], 0, data[i].length, res, curpos);
				}
				cobj.doFinal(res, curpos); 
				break;

			case block:
				switch (params.bulk_cipher_alg) {
				case bulk_3des:
					alg = "DESede";
					algmode = "DESede/CBC/NoPadding"; break;
				case bulk_aes128:
				case bulk_aes256:
					alg = "AES";
					algmode = "AES/CBC/NoPadding"; break;
				default:
					System.out.println("TLSState::decrypt: Algorithm not supported");
					return null;
				} 
				// init Crypto Object
				sk = new SecretKeySpec(keys.encKey, alg);
				cobj = Cipher.getInstance(algmode);
				rnd = new Random(System.currentTimeMillis());
				byte[] iv = new byte[params.record_iv_length];
				rnd.nextBytes(iv);
				IvParameterSpec paramSpec = new IvParameterSpec(iv);
				cobj.init(Cipher.ENCRYPT_MODE, sk, paramSpec);
				// create padding
				if(padding == null) {
					padlen = params.getMinPaddingSize(data);
					if(padlen != -1) 
						padding = createPadding(padlen, PaddingFill.FILL_PAD_LEN);
					else
						padding = new byte[0];
				}
				DebugInfo.printDump(4,padding,"padding");
				// encrypt data
				enclen = cobj.getOutputSize(datalen + padding.length);
				res = new byte[enclen + params.record_iv_length];
				System.arraycopy(iv, 0, res, 0, params.record_iv_length);
				for(i=0,curpos=0; i<data.length; i++){
					if(data[i] != null)
						curpos += cobj.update(data[i], 0, data[i].length, res, params.record_iv_length + curpos);
				}
				if(padding.length != 0)
					curpos += cobj.update(padding, 0, padding.length, res, params.record_iv_length + curpos);
				cobj.doFinal(res, params.record_iv_length + curpos); 
				break;

			case aead:
			default:
				break;
		}
		
		return res;
	}

	/*
	 * data is structure from rfc5246 (i.e. include IV, if it exist):
	 * 		case stream: GenericStreamCipher;
	 * 		case block:  GenericBlockCipher;
	 * 		case aead:   GenericAEADCipher; 
	 */
	public byte[] decrypt(byte[] data) throws Exception {
		byte[] res = null;
		SecretKeySpec sk;
		String alg, algmode;
		Cipher cobj;
		
		switch (params.cipher_type) {
			case stream:
				switch (params.bulk_cipher_alg) {
					case bulk_null:
						return data;
					case bulk_rc4:
						alg = "RC4";
						break;
					default:
						System.out.println("TLSState::decrypt: Algorithm not supported");
						return null;
				} 
				cobj = Cipher.getInstance(alg);
				sk = new SecretKeySpec(keys.encKey, alg);
				cobj.init(Cipher.DECRYPT_MODE, sk);
				res = cobj.doFinal(data); 
				break;

			case block:
				switch (params.bulk_cipher_alg) {
				case bulk_3des:
					alg = "DESede";
					algmode = "DESede/CBC/NoPadding"; break;
				case bulk_aes128:
				case bulk_aes256:
					alg = "AES";
					algmode = "AES/CBC/NoPadding"; break;
				default:
					System.out.println("TLSState::decrypt: Algorithm not supported");
					return null;
				} 
				sk = new SecretKeySpec(keys.encKey, alg);
				cobj = Cipher.getInstance(algmode);
				byte[] iv = new byte[params.record_iv_length];
				System.arraycopy(data, 0, iv, 0, params.record_iv_length);
				IvParameterSpec paramSpec = new IvParameterSpec(iv);
				cobj.init(Cipher.DECRYPT_MODE, sk, paramSpec);
				res = cobj.doFinal(data, params.record_iv_length, data.length - params.record_iv_length); 
				break;

			case aead:
			default:
				break;
		}
		
		return res;
	}
	
	public byte[] mac(byte[] data) throws Exception {
		Mac mobj;
		byte[] res = null;
		SecretKeySpec sk;
		String alg;
		
		if(data == null)
			return null;
		if(params.mac_alg == MACAlgorithm.mac_null)
			return new byte[0];
		
		if(keys.macKey != null) {
			switch (params.mac_alg) {
				case md5:
					if(keys.macKey.length != 16) return null;
					alg = "HmacMD5";
					break;
				case sha1:
					if(keys.macKey.length != 20) return null;
					alg = "HmacSHA1";
					break;
				case sha256:
					if(keys.macKey.length != 32) return null;
					alg = "HmacSHA256";
					break;
				default:
					System.out.println("TLSState::mac: Algorithm not supported");
					return null;
			}
			
			mobj = Mac.getInstance(alg);
			sk = new SecretKeySpec(keys.macKey, alg);
			mobj.init(sk);
			res = mobj.doFinal(data);
		}
		
		return res;
	}
	
	public byte[] mac(byte[][] data) throws Exception {
		Mac mobj;
		byte[] res = null;
		SecretKeySpec sk;
		String alg;
		
		if(data == null)
			return null;
		if(params.mac_alg == MACAlgorithm.mac_null)
			return new byte[0];
		
		if(keys.macKey != null) {
			switch (params.mac_alg) {
				case md5:
					if(keys.macKey.length != 16) return null;
					alg = "HmacMD5";
					break;
				case sha1:
					if(keys.macKey.length != 20) return null;
					alg = "HmacSHA1";
					break;
				case sha256:
					if(keys.macKey.length != 32) return null;
					alg = "HmacSHA256";
					break;
				default:
					System.out.println("TLSState::mac: Algorithm not supported");
					return null;
			}
			
			mobj = Mac.getInstance(alg);
			sk = new SecretKeySpec(keys.macKey, alg);
			mobj.init(sk);
			for(int i=0; i<data.length; i++) {
				if(data[i] != null)
					mobj.update(data[i]);
			}
			res = mobj.doFinal();
		}
		
		return res;
	}
	
	// include one byte for 'Padding_length' field
	public static byte[] createPadding(int len, PaddingFill fl) {
		byte[] res;
		
		if((len < 1) || (len > 256)) {
			System.out.println("TLSState::createPadding: Bad padding length");
			return null;
		}
		
		res = new byte[len];
		len--;
		
		for(int i=0;i<len;i++){
			switch (fl) {
				case FILL_PAD_LEN:
					res[i] = (byte)len; break;
				case FILL_PAD_ZERO:
					res[i] = 0; break;
				case FILL_PAD_NUM:
					res[i] = (byte)(i+1); break;
			}
		}
		res[len] = (byte)len;
		return res;
	}
	
	// include one byte for 'Padding_length' field
	public static boolean checkPadding(byte[] data) {
		boolean res = false;
		if(data != null) {
			int len = data.length - 1;
			for(int i=0;i<=len;i++){
				if(data[i] != (byte)len) 
					return false;
			}
			res = true;
		}
		return res;
	}
	
	public static boolean checkPadding(byte[] data, int offset) {
		boolean res = false;
		if(data != null && offset < data.length) {
			int len = data.length - 1 - offset;
			for(int i=0;i<=len;i++){
				if(data[offset + i] != (byte)len) 
					return false;
			}
			res = true;
		}
		return res;
	}
	
	public byte[] getKey(TLSStateKeyType type){
		byte[] key = null;

		if(keys != null)
			switch(type) {
				case MAC_KEY: key = keys.macKey; break;
				case ENC_KEY: key = keys.encKey; break;
				case FIXEDIV_KEY: key = keys.fixedIV; break;
			}
		
		return key;
	}

	public long getSequenceNumber(){
		return (keys != null) ? keys.sec_num : -1;
	}
	public void setSequenceNumber(long sec_num){
		keys.sec_num = sec_num;
	}
	public void incrementSequenceNumber(){
		keys.sec_num ++;
	}
	public void decrementSequenceNumber(){
		keys.sec_num --;
	}
		
	public void printDump(int debugLevel) {
		DebugInfo.printMessage(debugLevel,"Sequence number " + keys.sec_num);
		DebugInfo.printDump(debugLevel, keys.macKey, "macKey");
		DebugInfo.printDump(debugLevel, keys.encKey, "encKey");
		DebugInfo.printDump(debugLevel, keys.fixedIV, "fixedIV");
		getParams().printDump(debugLevel);
	}
}
